@@ -15,28 +15,20 @@
 """Jax transform lifting."""
 
 import collections
+from collections.abc import Callable, Iterable, Mapping, Sequence
+import contextlib
+import dataclasses
 import functools
-from typing import (
-    Any,
-    Callable,
-    Dict,
-    Iterable,
-    List,
-    Mapping,
-    Optional,
-    Sequence,
-    Tuple,
-    TypeVar,
-    Union,
-)
+import threading
+from typing import Any, Generic, TypeVar
 import warnings
 
 from flax import traceback_util
 from flax.typing import (
-  In,
-  Out,
-  InOutAxis,
-  InOutScanAxis,
+    In,
+    InOutAxis,
+    InOutScanAxis,
+    Out,
 )
 import jax
 from jax import random
@@ -44,21 +36,42 @@ from jax import random
 from . import axes_scan, meta
 from .frozen_dict import freeze, unfreeze
 from .scope import (
-  CollectionFilter,
-  DenyList,  # pylint: disable=g-multiple-import
-  Filter,
-  PRNGSequenceFilter,
-  Scope,
-  group_collections,
-  in_filter,
-  intersect_filters,
-  is_filter_empty,
-  subtract_filters,
-  union_filters,
+    CollectionFilter,
+    DenyList,  # pylint: disable=g-multiple-import
+    Filter,
+    LazyRng,
+    PRNGSequenceFilter,
+    Scope,
+    group_collections,
+    in_filter,
+    intersect_filters,
+    is_filter_empty,
+    subtract_filters,
+    union_filters,
 )
 
 traceback_util.register_exclusion(__file__)
 
+A = TypeVar('A')
+
+
+@dataclasses.dataclass
+class TransformContext(Generic[A], threading.local):
+  """Context for a transform."""
+
+  stack: list[A] = dataclasses.field(default_factory=list)
+
+  @contextlib.contextmanager
+  def push(self, a: A):
+    self.stack.append(a)
+    try:
+      yield
+    finally:
+      self.stack.pop()
+
+  def get(self) -> A:
+    return self.stack[-1]
+
 
 def tree_map_rngs(fn, tree):
   """Needed for mapping JAX random.* functions over PRNGKey leaves."""
@@ -162,7 +175,7 @@ def _partial_pack(
     inner_rng_counters.append(rng_counters)
   rng_groups_xs_t = _transpose(rng_groups_xs)
 
-  inner_scopes: List[Scope] = []
+  inner_scopes: list[Scope] = []
 
   def scope_fn(
       variable_groups_xs_t,
@@ -439,7 +452,7 @@ def vjp(
   vjp_variables: CollectionFilter = 'params',
   variables: CollectionFilter = True,
   rngs: PRNGSequenceFilter = True,
-) -> Union[Tuple[Any, Callable[..., Any]], Tuple[Any, Callable[..., Any], Any]]:
+) -> tuple[Any, Callable[..., Any]] | tuple[Any, Callable[..., Any], Any]:
   """A lifted version of ``jax.vjp``.
 
   See ``jax.vjp`` for the unlifted vector-Jacobian product (backward gradient).
@@ -538,7 +551,7 @@ def value_and_grad(
   reduce_axes=(),
   variables: CollectionFilter = True,
   rngs: PRNGSequenceFilter = True,
-) -> Union[Tuple[Any, Callable[..., Any]], Tuple[Any, Callable[..., Any], Any]]:
+) -> tuple[Any, Callable[..., Any]] | tuple[Any, Callable[..., Any], Any]:
   """A limited lifted version of ``jax.value_and_grad``.
 
   See ``jax.value_and_grad`` for the unlifted reverse mode gradient.
@@ -631,7 +644,7 @@ def jvp(
   variable_tangents,
   variables: CollectionFilter = True,
   rngs: PRNGSequenceFilter = True,
-) -> Tuple[Any, Any]:
+) -> tuple[Any, Any]:
   """A lifted version of ``jax.jvp``.
 
   See ``jax.jvp`` for the unlifted Jacobian-vector product (forward gradient).
@@ -720,10 +733,10 @@ def vmap(
   split_rngs: Mapping[PRNGSequenceFilter, bool],
   in_axes=0,
   out_axes=0,
-  axis_size: Optional[int] = None,
-  axis_name: Optional[str] = None,
-  spmd_axis_name: Optional[str] = None,
-  metadata_params: Dict[Any, Any] = {},
+  axis_size: int | None = None,
+  axis_name: str | None = None,
+  spmd_axis_name: str | None = None,
+  metadata_params: dict[Any, Any] = {},
 ) -> Callable[..., Any]:
   """A lifted version of ``jax.vmap``.
 
@@ -872,12 +885,12 @@ def scan(
   split_rngs: Mapping[PRNGSequenceFilter, bool] = {},
   in_axes=0,
   out_axes=0,
-  length: Optional[int] = None,
+  length: int | None = None,
   reverse: bool = False,
   unroll: int = 1,
   _split_transpose: bool = False,
-  data_transform: Optional[Callable[..., Any]] = None,
-  metadata_params: Dict[Any, Any] = {},
+  data_transform: Callable[..., Any] | None = None,
+  metadata_params: dict[Any, Any] = {},
 ) -> Callable[..., Any]:
   """A lifted version of ``jax.lax.scan``.
 
@@ -1416,20 +1429,20 @@ def checkpoint(
   rngs: PRNGSequenceFilter = True,
   concrete: bool = False,
   prevent_cse: bool = True,
-  static_argnums: Union[int, Tuple[int, ...]] = (),
-  policy: Optional[Callable[..., bool]] = None,
+  static_argnums: int | tuple[int, ...] = (),
+  policy: Callable[..., bool] | None = None,
 ) -> Callable[..., Any]:
   """Lifted version of ``jax.checkpoint``.
 
   This function is aliased to ``lift.remat`` just like ``jax.remat``.
 
   Args:
-    fn: scope function for which intermediate computations should be
-    re-computed when computing gradients.
+    fn: scope function for which intermediate computations should be re-computed
+      when computing gradients.
     variables: The variable collections that are lifted. By default all
       collections are lifted.
-    rngs: The PRNG sequences that are lifted. By default all PRNG sequences
-      are lifted.
+    rngs: The PRNG sequences that are lifted. By default all PRNG sequences are
+      lifted.
     concrete: Optional, boolean indicating whether ``fun`` may involve
       value-dependent Python control flow (default ``False``). Support for such
       control flow is optional, and disabled by default, because in some
@@ -1448,6 +1461,7 @@ def checkpoint(
       arguments as static can avoid ConcretizationTypeErrors when tracing, but
       at the cost of more retracing overheads.
     policy: Experimental checkpoint policy, see ``jax.checkpoint``.
+
   Returns:
     A wrapped version of ``fn``. When computing gradients intermediate
     computations will be re-computed when computing gradients.
@@ -1502,11 +1516,11 @@ def jit(
     fn: Callable[..., Any],
     variables: CollectionFilter = True,
     rngs: PRNGSequenceFilter = True,
-    static_argnums: Union[int, Iterable[int]] = (),
-    static_argnames: Union[str, Iterable[str]] = (),
-    donate_argnums: Union[int, Iterable[int]] = (),
+    static_argnums: int | Iterable[int] = (),
+    static_argnames: str | Iterable[str] = (),
+    donate_argnums: int | Iterable[int] = (),
     device=None,
-    backend: Union[str, None] = None,
+    backend: str | None = None,
 ) -> Callable[..., Any]:
   """Lifted version of ``jax.jit``.
 
@@ -1562,8 +1576,7 @@ def jit(
   # Close over scope_fn & repack_fn to avoid recompilation
   # this is impure but we use the fingerprint arg to differentiate between cases
   # where scope_fn or repack_fn actually produce non-identical results.
-  scope_fn = None  # type: Optional[Callable]
-  repack_fn = None  # type: Optional[Callable]
+  jit_context = TransformContext[tuple[Callable, Callable]]()
 
   @functools.partial(
       jax.jit,
@@ -1575,33 +1588,37 @@ def jit(
   )
   @functools.wraps(fn)
   def jitted(fingerprint, variable_groups, rng_groups, *args, **kwargs):
-    nonlocal scope_fn, repack_fn
+    scope_fn, repack_fn = jit_context.get()
     hash_key = fingerprint[1]
     # fingerprint is only used to differentiate the cache signature
-    del fingerprint
+    # del fingerprint
     scope = scope_fn(variable_groups, rng_groups)  # pylint: disable=not-callable
     y = fn(scope, hash_key, *args, **kwargs)
     return y, repack_fn(scope)  # pylint: disable=not-callable
 
   def inner(
-      scope_fun,
-      repack_fun,
+      scope_fn,
+      repack_fn,
       variable_groups,
       rng_groups,
       module_hash_key,
       *args,
       **kwargs,
   ):
-    nonlocal scope_fn, repack_fn
-    try:
-      scope_fn = scope_fun
-      repack_fn = repack_fun
-      scopes = jax.tree_util.tree_leaves(scope_fn(variable_groups, rng_groups))
+    with jit_context.push((scope_fn, repack_fn)):
+      scopes: list[Scope] = jax.tree_util.tree_leaves(
+          scope_fn(variable_groups, rng_groups)
+      )
       mutable = tuple(_hashable_filter(scope.mutable) for scope in scopes)
+
+      rng_groups = jax.tree.map(
+          lambda x: x.fold() if isinstance(x, LazyRng) else x,
+          rng_groups,
+          is_leaf=lambda x: isinstance(x, LazyRng),
+      )
+
       fingerprint = (mutable, module_hash_key)
       return jitted(fingerprint, variable_groups, rng_groups, *args, **kwargs)
-    finally:
-      scope_fn, repack_fn = None, None
 
   return pack(
       inner, (variables,), (variables,), (rngs,), name='jit', enable_kwargs=True
@@ -1611,7 +1628,7 @@ def jit(
 def remat_scan(
   body_fn: Callable[..., Any],
   lengths: Sequence[int],
-  policy: Optional[Callable[..., bool]] = None,
+  policy: Callable[..., bool] | None = None,
   variable_broadcast: CollectionFilter = False,
   variable_carry: CollectionFilter = False,
   variable_axes: Mapping[CollectionFilter, InOutScanAxis] = {True: 0},
