@@ -87,6 +87,7 @@ from .attention import (
 from .batch_apply import BatchApply as BatchApply
 from .combinators import Sequential as Sequential
 from .fp8_ops import Fp8DotGeneralOp as Fp8DotGeneralOp
+from .fp8_ops import NANOOFp8DotGeneralOp as NANOOFp8DotGeneralOp
 from .initializers import (
     ones_init as ones_init,
     ones as ones,
@@ -116,6 +117,7 @@ from .module import (
     merge_param as merge_param,
     nowrap as nowrap,
     override_named_call as override_named_call,
+    share_scope as share_scope,
 )
 from .normalization import (
     BatchNorm as BatchNorm,
