@@ -27,13 +27,13 @@
 # limitations under the License.
 """APIs for handling dtypes in Linen Modules."""
 
-from typing import Any, List, Optional
+from typing import Any
 from flax.typing import Dtype
 from jax import numpy as jnp
 
 
 def canonicalize_dtype(
-  *args, dtype: Optional[Dtype] = None, inexact: bool = True
+  *args, dtype: Dtype | None = None, inexact: bool = True
 ) -> Dtype:
   """Canonicalize an optional dtype to the definitive dtype.
 
@@ -64,7 +64,7 @@ def canonicalize_dtype(
   return dtype
 
 
-def promote_dtype(*args, dtype=None, inexact=True) -> List[Any]:
+def promote_dtype(*args, dtype=None, inexact=True) -> list[Any]:
   """ "Promotes input arguments to a specified or inferred dtype.
 
   All args are cast to the same dtype. See ``canonicalize_dtype`` for how
