@@ -42,17 +42,10 @@ import functools
 import inspect
 from typing import (
   Any,
-  Callable,
-  Dict,
-  Iterable,
-  Mapping,
-  Optional,
-  Sequence,
-  Tuple,
-  Type,
   TypeVar,
   Union,
 )
+from collections.abc import Callable, Iterable, Mapping, Sequence
 
 from flax import core
 from flax import errors, struct, traceback_util
@@ -109,8 +102,8 @@ class VariablePlaceholder:
 class InstancePlaceholder:
   """Marks module instances in a JAX-compatible way when lifting arguments."""
 
-  cls: Type[Any] = struct.field(pytree_node=False)
-  attrs: Dict[Any, Any] = struct.field(pytree_node=False)
+  cls: type[Any] = struct.field(pytree_node=False)
+  attrs: dict[Any, Any] = struct.field(pytree_node=False)
   id: int = struct.field(pytree_node=False)
 
 
@@ -159,7 +152,7 @@ def get_module_scopes(module, args=None, kwargs=None):
     VariablePlaceholders and Module instances replaced with InstancePlaceholders
     that are compatible with jax functions.
   """
-  scopes = []
+  scopes: list[Scope] = []
   refs = {}
 
   # Gather scopes associated with Variables and Module instances passed as
@@ -491,6 +484,8 @@ def _fingerprint_recursive(
 
   if isinstance(obj, str):
     return obj
+  elif hasattr(obj, '__fn_or_cls__'):  # support PaxConfig objects
+    return _fingerprint_recursive(obj.__fn_or_cls__, path, seen_modules)
   elif isinstance(obj, Module):
     fingerprint: Any
     if obj._id in seen_modules:
@@ -569,7 +564,7 @@ def _check_field_is_hashable(path: tuple[str, ...], x: Any):
     raise ValueError(f"Value at '{path_name}' is not hashable: {e}") from e
 
 
-def decorator_lift_transform_jit(class_fn, **trafo_kwargs):
+def decorator_lift_transform_cached(transform, class_fn, **trafo_kwargs):
   """Decorator for lifted transform.
 
   Similar to `decorator_lift_transform` but specialized for `jit`, it reuses the
@@ -579,7 +574,6 @@ def decorator_lift_transform_jit(class_fn, **trafo_kwargs):
   # Due to the ordering of method decorators, we must wrap the class_fn
   # with the module state management wrapper first to maintain Module state
   # correctly.
-  transform = lift.jit
   multi_scope = True
 
   if isinstance(class_fn, tuple):
@@ -615,8 +609,10 @@ def decorator_lift_transform_jit(class_fn, **trafo_kwargs):
       return res
 
     core_fns = [
-      functools.partial(core_fn, prewrapped_fn, class_fn)
-      for prewrapped_fn, class_fn in zip(prewrapped_fns, class_fns)
+        functools.wraps(class_fn)(
+            functools.partial(core_fn, prewrapped_fn, class_fn)
+        )
+        for prewrapped_fn, class_fn in zip(prewrapped_fns, class_fns)
     ]
 
     # here we apply the given lifting transform to the scope-ingesting fn
@@ -624,6 +620,7 @@ def decorator_lift_transform_jit(class_fn, **trafo_kwargs):
       trafo_fn = transform(*core_fns, **trafo_kwargs)
 
     module_scopes, args, kwargs = get_module_scopes(self, args, kwargs)
+
     if not multi_scope:
       if len(module_scopes) != 1:
         # TODO(levskaya): transforms like jvp & vjp have args that follow the
@@ -645,11 +642,12 @@ def decorator_lift_transform_jit(class_fn, **trafo_kwargs):
   return wrapped_fn
 
 
-def module_class_lift_transform_jit(module_class, methods=None, **trafo_kwargs):
+def module_class_lift_transform_cached(
+    transform, module_class, methods=None, **trafo_kwargs
+):
   """Module class lift transform."""
   # TODO(marcvanzee): Improve docstrings (#1977).
   # TODO(levskaya): find nicer argument convention for multi-method case?
-  transform = lift.jit
   trafo_args = ()
 
   # Prepare per-method transform args, kwargs.
@@ -739,7 +737,7 @@ def module_class_lift_transform_jit(module_class, methods=None, **trafo_kwargs):
 # Utility to wrap a class or to use as decorator in def of class method.
 # -----------------------------------------------------------------------------
 
-TransformTarget = Union[Type[Module], Callable[..., Any]]
+TransformTarget = Union[type[Module], Callable[..., Any]]
 Target = TypeVar('Target', bound=TransformTarget)
 
 
@@ -770,9 +768,27 @@ def lift_transform(
     raise errors.TransformTargetError(target)
 
 
+def lift_transform_cached(
+    transform, target, *trafo_args, methods=None, **trafo_kwargs
+):
+  """Applies to class or as a decorator on class fns."""
+  # TODO(marcvanzee): Improve docstrings (#1977).
+  if _is_module_class(target):
+    return module_class_lift_transform_cached(
+        transform, target, *trafo_args, methods=methods, **trafo_kwargs
+    )
+  # we presume this is being used as a function decorator in class definition
+  elif callable(target) and not isinstance(target, Module):
+    return decorator_lift_transform_cached(
+        transform, target, *trafo_args, **trafo_kwargs
+    )
+  else:
+    raise errors.TransformTargetError(target)
+
+
 def lift_direct_transform(
   transform: Callable[..., Any],
-  targets: Tuple[Callable[..., Any], ...],
+  targets: tuple[Callable[..., Any], ...],
   mdl: Module,
   *args,
   multi_scope=True,
@@ -803,9 +819,9 @@ def vmap(
   split_rngs: Mapping[PRNGSequenceFilter, bool] = FrozenDict(),
   in_axes=0,
   out_axes=0,
-  axis_size: Optional[int] = None,
-  axis_name: Optional[str] = None,
-  spmd_axis_name: Optional[str] = None,
+  axis_size: int | None = None,
+  axis_name: str | None = None,
+  spmd_axis_name: str | None = None,
   metadata_params: Mapping[Any, Any] = {},
   methods=None,
 ) -> Target:
@@ -894,11 +910,11 @@ def jit(
   target: Target,
   variables: CollectionFilter = True,
   rngs: PRNGSequenceFilter = True,
-  static_argnums: Union[int, Iterable[int]] = (),
-  static_argnames: Union[str, Iterable[str]] = (),
-  donate_argnums: Union[int, Iterable[int]] = (),
+  static_argnums: int | Iterable[int] = (),
+  static_argnames: str | Iterable[str] = (),
+  donate_argnums: int | Iterable[int] = (),
   device=None,
-  backend: Union[str, None] = None,
+  backend: str | None = None,
   methods=None,
 ) -> Target:
   """Lifted version of ``jax.jit``.
@@ -946,8 +962,8 @@ def jit(
     A wrapped version of target, set up for just-in-time compilation.
   """
   # TODO(marcvanzee): Improve docstrings (#1977).
-  if _is_module_class(target):
-    return module_class_lift_transform_jit(
+  return lift_transform_cached(
+      lift.jit,
       target,
       variables=variables,
       rngs=rngs,
@@ -957,21 +973,7 @@ def jit(
       device=device,
       backend=backend,
       methods=methods,
-    )
-  # we presume this is being used as a function decorator in class definition
-  elif callable(target) and not isinstance(target, Module):
-    return decorator_lift_transform_jit(
-      target,
-      variables=variables,
-      rngs=rngs,
-      static_argnums=static_argnums,
-      static_argnames=static_argnames,
-      donate_argnums=donate_argnums,
-      device=device,
-      backend=backend,
-    )
-  else:
-    raise errors.TransformTargetError(target)
+  )
 
 
 def checkpoint(
@@ -980,8 +982,8 @@ def checkpoint(
   rngs: PRNGSequenceFilter = True,
   concrete: bool = False,
   prevent_cse: bool = True,
-  static_argnums: Union[int, Tuple[int, ...]] = (),
-  policy: Optional[Callable[..., bool]] = None,
+  static_argnums: int | tuple[int, ...] = (),
+  policy: Callable[..., bool] | None = None,
   methods=None,
 ) -> Target:
   """Lifted version of ``jax.checkpoint``.
@@ -1049,15 +1051,15 @@ def checkpoint(
   # lifted function
   static_argnums = jax.tree_util.tree_map(lambda x: x - 1, static_argnums)
   return lift_transform(
-    lift.checkpoint,
-    target,
-    variables=variables,
-    rngs=rngs,
-    concrete=concrete,
-    static_argnums=static_argnums,
-    prevent_cse=prevent_cse,
-    policy=policy,
-    methods=methods,
+      lift.checkpoint,
+      target,
+      variables=variables,
+      rngs=rngs,
+      concrete=concrete,
+      static_argnums=static_argnums,
+      prevent_cse=prevent_cse,
+      policy=policy,
+      methods=methods,
   )
 
 
@@ -1066,8 +1068,8 @@ remat = checkpoint
 
 def remat_scan(
   target: Target,
-  lengths: Optional[Sequence[int]] = (),
-  policy: Optional[Callable[..., bool]] = None,
+  lengths: Sequence[int] | None = (),
+  policy: Callable[..., bool] | None = None,
   variable_broadcast: CollectionFilter = False,
   variable_carry: CollectionFilter = False,
   variable_axes: Mapping[CollectionFilter, InOutScanAxis] = FrozenDict(
@@ -1138,10 +1140,10 @@ def scan(
   split_rngs: Mapping[PRNGSequenceFilter, bool] = FrozenDict(),
   in_axes=0,
   out_axes=0,
-  length: Optional[int] = None,
+  length: int | None = None,
   reverse: bool = False,
   unroll: int = 1,
-  data_transform: Optional[Callable[..., Any]] = None,
+  data_transform: Callable[..., Any] | None = None,
   metadata_params: Mapping[Any, Any] = {},
   methods=None,
   _split_transpose: bool = False,
@@ -1659,7 +1661,7 @@ def jvp(
   variable_tangents,
   variables: CollectionFilter = True,
   rngs: PRNGSequenceFilter = True,
-) -> Union[Tuple[Any, Callable[..., Any]], Tuple[Any, Callable[..., Any], Any]]:
+) -> tuple[Any, Callable[..., Any]] | tuple[Any, Callable[..., Any], Any]:
   """A lifted version of ``jax.jvp``.
 
   See ``jax.jvp`` for the unlifted Jacobian-vector product (forward gradient).
@@ -2110,7 +2112,7 @@ def named_call(class_fn, force=True):
 def add_metadata_axis(
   target: Target,
   variable_axes: Mapping[CollectionFilter, InOutAxis] = FrozenDict(),
-  metadata_params: Dict[Any, Any] = {},
+  metadata_params: dict[Any, Any] = {},
 ) -> Target:
   """A helper to manipulate boxed axis metadata.
 
