@@ -14,18 +14,18 @@
 
 from __future__ import annotations
 
-from collections import defaultdict
+import contextlib
 import dataclasses
 import enum
 import functools
 import threading
 import typing as tp
-from copy import deepcopy
 
 import jax
 import numpy as np
 import typing_extensions as tpe
 
+from flax.core.frozen_dict import FrozenDict
 from flax.nnx.nnx import filterlib, reprlib
 from flax.nnx.nnx.proxy_caller import (
   ApplyCaller,
@@ -59,16 +59,6 @@ def is_state_leaf(x: tp.Any) -> tpe.TypeGuard[StateLeaf]:
   return isinstance(x, (VariableState, np.ndarray, jax.Array))
 
 
-@dataclasses.dataclass
-class GraphContext(threading.local):
-  update_context_stacks: defaultdict[str, list[UpdateContext]] = (
-    dataclasses.field(default_factory=lambda: defaultdict(list))
-  )
-
-
-GRAPH_CONTEXT = GraphContext()
-
-
 def is_node_leaf(x: tp.Any) -> tpe.TypeGuard[StateLeaf]:
   return isinstance(x, (Variable, np.ndarray, jax.Array))
 
@@ -157,7 +147,7 @@ NodeImpl = tp.Union[
 ]
 
 
-_node_impl_for_type: dict[type, 'NodeImpl[tp.Any, tp.Any, tp.Any]'] = {}
+_node_impl_for_type: dict[type, NodeImpl[tp.Any, tp.Any, tp.Any]] = {}
 
 
 def register_graph_node_type(
@@ -241,15 +231,49 @@ class _HashableMapping(tp.Mapping[HA, HB], tp.Hashable):
     return repr(self._mapping)
 
 
+class GraphDef(tp.Generic[Node]):
+  type: type[Node]
+  index: int
+
+
 @dataclasses.dataclass(frozen=True, repr=False)
-class NodeDef(tp.Generic[Node], reprlib.Representable):
+class NodeRef(GraphDef[Node], reprlib.Representable):
+  type: type[Node]
+  index: int
+
+  def __nnx_repr__(self):
+    yield reprlib.Object(type=type(self))
+    yield reprlib.Attr('type', self.type.__name__)
+    yield reprlib.Attr('index', self.index)
+
+  def __penzai_repr__(self, path, subtree_renderer):
+    from penzai.treescope import repr_lib as pz_repr_lib  # type: ignore[import-not-found,import-untyped]
+
+    return pz_repr_lib.render_object_constructor(
+      object_type=type(self),
+      attributes={'type': self.type, 'index': self.index},
+      path=path,
+      subtree_renderer=subtree_renderer,
+    )
+
+
+jax.tree_util.register_static(NodeRef)
+
+
+@dataclasses.dataclass(frozen=True, repr=False)
+class NodeDef(GraphDef[Node], reprlib.Representable):
+  """A dataclass that denotes the tree structure of a
+  :class:`Module`. A ``GraphDef`` can be generated by either
+  calling :func:`split` or :func:`graphdef` on the :class:`Module`."""
+
   type: tp.Type[Node]
   index: int
   attributes: tuple[Key, ...]
-  subgraphs: _HashableMapping[Key, tp.Union['NodeDef[tp.Any]', Index]]
+  subgraphs: _HashableMapping[Key, NodeDef[tp.Any] | NodeRef[tp.Any]]
   static_fields: _HashableMapping[Key, tp.Any]
-  leaves: _HashableMapping[Key, Index | None]
+  leaves: _HashableMapping[Key, NodeRef[tp.Any] | None]
   metadata: tp.Any
+  index_mapping: FrozenDict[Index, Index] | None
 
   @classmethod
   def create(
@@ -257,10 +281,11 @@ class NodeDef(tp.Generic[Node], reprlib.Representable):
     type: tp.Type[Node],
     index: int,
     attributes: tuple[Key, ...],
-    subgraphs: tp.Iterable[tuple[Key, tp.Union['NodeDef[tp.Any]', Index]]],
+    subgraphs: tp.Iterable[tuple[Key, NodeDef[tp.Any] | NodeRef[tp.Any]]],
     static_fields: tp.Iterable[tuple[Key, tp.Any]],
-    leaves: tp.Iterable[tuple[Key, Index | None]],
+    leaves: tp.Iterable[tuple[Key, NodeRef[tp.Any] | None]],
     metadata: tp.Any,
+    index_mapping: tp.Mapping[Index, Index] | None,
   ):
     return cls(
       type=type,
@@ -270,6 +295,9 @@ class NodeDef(tp.Generic[Node], reprlib.Representable):
       static_fields=_HashableMapping(static_fields),
       leaves=_HashableMapping(leaves),
       metadata=metadata,
+      index_mapping=FrozenDict(index_mapping)
+      if index_mapping is not None
+      else None,
     )
 
   def __nnx_repr__(self):
@@ -284,10 +312,16 @@ class NodeDef(tp.Generic[Node], reprlib.Representable):
     )
     yield reprlib.Attr('leaves', reprlib.PrettyMapping(self.leaves))
     yield reprlib.Attr('metadata', self.metadata)
+    yield reprlib.Attr(
+      'index_mapping',
+      reprlib.PrettyMapping(self.index_mapping)
+      if self.index_mapping is not None
+      else None,
+    )
 
-  def __penzai_repr__(self, path, subtree_renderer):
-    from penzai.treescope import repr_lib as pz_repr_lib  # type: ignore[import-not-found,import-untyped]
-    return pz_repr_lib.render_object_constructor(
+  def __treescope_repr__(self, path, subtree_renderer):
+    import treescope  # type: ignore[import-not-found,import-untyped]
+    return treescope.repr_lib.render_object_constructor(
         object_type=type(self),
         attributes={
             'type': self.type,
@@ -302,48 +336,9 @@ class NodeDef(tp.Generic[Node], reprlib.Representable):
         subtree_renderer=subtree_renderer,
     )
 
-
-@dataclasses.dataclass(frozen=True, repr=False)
-class GraphDef(tp.Generic[Node], reprlib.Representable):
-  """A dataclass that denotes the tree structure of a
-  :class:`Module`. A ``GraphDef`` can be generated by either
-  calling :func:`split` or :func:`graphdef` on the :class:`Module`."""
-
-  nodedef: NodeDef[Node]
-  index_mapping: dict[Index, Index] | None
-
-  def __nnx_repr__(self):
-    yield reprlib.Object(type=type(self))
-
-    yield reprlib.Attr('nodedef', self.nodedef)
-    yield reprlib.Attr('index_mapping', self.index_mapping)
-
-  def __penzai_repr__(self, path, subtree_renderer):
-    from penzai.treescope import repr_lib as pz_repr_lib  # type: ignore[import-not-found,import-untyped]
-    return pz_repr_lib.render_object_constructor(
-        object_type=type(self),
-        attributes={
-            'nodedef': self.nodedef,
-            'index_mapping': self.index_mapping,
-        },
-        path=path,
-        subtree_renderer=subtree_renderer,
-    )
-
-  def __deepcopy__(self, memo=None):
-    nodedef = deepcopy(self.nodedef, memo)
-    index_mapping = deepcopy(self.index_mapping, memo)
-    return GraphDef(nodedef, index_mapping)
-
-  def __hash__(self):
-    return hash(self.nodedef)
-
-  def __eq__(self, other):
-    return isinstance(other, GraphDef) and self.nodedef == other.nodedef
-
   def apply(
     self, state: GraphState, *states: GraphState
-  ) -> ApplyCaller[tuple['GraphDef[Node]', GraphState]]:
+  ) -> ApplyCaller[tuple[GraphDef[Node], GraphState]]:
     accessor = DelayedAccessor()
 
     def _apply(
@@ -352,85 +347,71 @@ class GraphDef(tp.Generic[Node], reprlib.Representable):
       module = merge(self, state, *states)
       fn = accessor(module)
       out = fn(*args, **kwargs)
-      return out, flatten(module)[:2]
+      return out, flatten(module)
 
     return CallableProxy(_apply, accessor)  # type: ignore
 
-def _graphdef_flatten(graphdef: GraphDef[Node]):
-  # refmap is opaque, we don't propagate it
-  static = (graphdef.nodedef, graphdef.index_mapping)
-  return (), static
 
+jax.tree_util.register_static(NodeDef)
 
-def _graphdef_unflatten(
-  static: tuple[NodeDef[Node], dict[Index, Index] | None], _nodes: tuple[()]
-):
-  nodedef, index_mapping = static
-  return GraphDef(nodedef, index_mapping)
-
-
-jax.tree_util.register_pytree_node(
-  GraphDef,
-  _graphdef_flatten,
-  _graphdef_unflatten,
-)
+PureState = tuple[GraphDef[A], GraphState]
 
 
 def flatten(
-  x: Node,
-  /,
-  *,
-  idxmap: dict[Index, tp.Any] | None = None,
-) -> tuple[GraphDef[Node], GraphState, RefMap[tp.Any, Index]]:
-  refmap = RefMap[tp.Any, Index]()
+  node: Node, /, ref_index: RefMap[tp.Any, Index] | None = None
+) -> tuple[GraphDef[Node], GraphState]:
+  """Flattens a graph node into a (graphdef, state) pair.
+
+  Args:
+    x: A graph node.
+    ref_index: A mapping from nodes to indexes, defaults to None. If not provided, a new
+      empty dictionary is created. This argument can be used to flatten a sequence of graph
+      nodes that share references.
+  """
+  if ref_index is None:
+    ref_index = RefMap()
   flat_state: dict[PathParts, StateLeaf] = {}
-  nodedef = _graph_flatten((), refmap, flat_state, x)
-  assert not isinstance(nodedef, int)
-  if idxmap is not None:
-    index_to_index = compose_mapping(idxmap, refmap)
-  else:
-    index_to_index = None
-  graphdef = GraphDef(nodedef, index_to_index)
-  return graphdef, GraphState.from_flat_path(flat_state), refmap
+  graphdef = _graph_flatten((), ref_index, flat_state, node)
+  return graphdef, GraphState.from_flat_path(flat_state)
 
 
 def _graph_flatten(
   path: PathParts,
-  refmap: RefMap[tp.Any, Index],
+  ref_index: RefMap[tp.Any, Index],
   flat_state: dict[PathParts, StateLeaf],
   node: Node,
-) -> NodeDef[Node] | int:
+) -> NodeDef[Node] | NodeRef:
   if not is_node(node):
     raise RuntimeError(f'Unsupported type: {type(node)}, this is a bug.')
 
-  if node in refmap:
-    return refmap[node]
+  if node in ref_index:
+    return NodeRef(type(node), ref_index[node])
 
   node_impl = get_node_impl(node)
 
   # only cache graph nodes
   if isinstance(node_impl, GraphNodeImpl):
-    index = len(refmap)
-    refmap[node] = index
+    index = len(ref_index)
+    ref_index[node] = index
   else:
     index = -1
 
-  subgraphs: list[tuple[Key, tp.Union[NodeDef[Node], Index]]] = []
+  subgraphs: list[tuple[Key, NodeDef[Node] | NodeRef]] = []
   static_fields: list[tuple[Key, tp.Any]] = []
-  leaves: list[tuple[Key, Index | None]] = []
+  leaves: list[tuple[Key, NodeRef | None]] = []
 
   values, metadata = node_impl.flatten(node)
   for key, value in values:
     if is_node(value):
-      nodedef = _graph_flatten((*path, key), refmap, flat_state, value)
+      nodedef = _graph_flatten((*path, key), ref_index, flat_state, value)
       subgraphs.append((key, nodedef))
     elif isinstance(value, Variable):
-      if value in refmap:
-        leaves.append((key, refmap[value]))
+      if value in ref_index:
+        leaves.append((key, NodeRef(type(value), ref_index[value])))
       else:
         flat_state[(*path, key)] = value.to_state()
-        variable_index = refmap[value] = len(refmap)
-        leaves.append((key, variable_index))
+        variable_index = ref_index[value] = len(ref_index)
+        leaves.append((key, NodeRef(type(value), variable_index)))
     elif is_state_leaf(value):
       flat_state[(*path, key)] = value
       leaves.append((key, None))
@@ -445,6 +426,7 @@ def _graph_flatten(
     static_fields=static_fields,
     leaves=leaves,
     metadata=metadata,
+    index_mapping=None,
   )
   return nodedef
 
@@ -454,53 +436,59 @@ def unflatten(
   state: GraphState,
   /,
   *,
-  idxmap: dict[Index, tp.Any] | None = None,
-) -> tuple[Node, dict[Index, tp.Any]]:
+  index_ref: dict[Index, tp.Any] | None = None,
+  index_ref_cache: dict[Index, tp.Any] | None = None,
+) -> Node:
   """Unflattens a graphdef into a node with the given state.
 
   Args:
-    graphdef: A NodeDef instance.
+    graphdef: A GraphDef instance.
     state: A State instance.
-    ref_cache: A mapping from indexes to existing nodes that can be reused.
+    index_ref: A mapping from indexes to nodes references found during the graph
+      traversal, defaults to None. If not provided, a new empty dictionary is
+      created. This argument can be used to unflatten a sequence of (graphdef, state)
+      pairs that share the same index space.
+    index_ref_cache: A mapping from indexes to existing nodes that can be reused.
       When an reference is reused, ``GraphNodeImpl.clear`` is called to leave the
       object in an empty state and then filled by the unflatten process, as a result
       existing graph nodes are mutated to have the new content/topology
       specified by the graphdef.
   """
-  index_to_ref: dict[Index, tp.Any] = {}
+  if index_ref is None:
+    index_ref = {}
+  assert isinstance(graphdef, (NodeDef, NodeRef))
   node = _graph_unflatten(
-    graphdef.nodedef, state.raw_mapping, index_to_ref, idxmap
+    graphdef, state.raw_mapping, index_ref, index_ref_cache
   )
-  return node, index_to_ref
-
+  return node
 
 def _graph_unflatten(
-  nodedef: tp.Union[NodeDef[Node], int],
+  nodedef: NodeDef[Node] | NodeRef[Node],
   state: tp.Mapping[Key, StateLeaf | tp.Mapping[Key, tp.Any]],
-  index_to_ref: dict[Index, tp.Any],
-  idxmap: dict[Index, tp.Any] | None,
+  index_ref: dict[Index, tp.Any],
+  index_ref_cache: dict[Index, tp.Any] | None,
 ) -> Node:
   """Recursive helper for graph_unflatten.
 
   Args:
-    nodedef: A NodeDef instance or an index to a node in the cache.
+    nodedef: A GraphDef instance or an index to a node in the cache.
     state: A mapping from attribute names to variables or subgraphs.
     index_to_ref: A mapping from indexes to nodes that have been traversed.
       If a node is already in the cache, it won't be traversed again.
-    ref_cache: A mapping from indexes to existing nodes that can be reused.
+    index_ref_cache: A mapping from indexes to existing nodes that can be reused.
       When an reference is reused, ``GraphNodeImpl.clear`` is called to leave the
       object in an empty state and then filled by the unflatten process, as a result
       existing graph nodes are mutated to have the new content/topology
       specified by the nodedef.
   """
-  if isinstance(nodedef, int):
-    return index_to_ref[nodedef]
+  if isinstance(nodedef, NodeRef):
+    return index_ref[nodedef.index]
 
   if not is_node_type(nodedef.type):
     raise RuntimeError(f'Unsupported type: {nodedef.type}, this is a bug.')
 
-  if nodedef.index in index_to_ref:
-    raise RuntimeError(f'NodeDef index {nodedef.index} already used.')
+  if nodedef.index in index_ref:
+    raise RuntimeError(f'GraphDef index {nodedef.index} already used.')
 
   node_impl = get_node_impl_for_type(nodedef.type)
 
@@ -523,9 +511,9 @@ def _graph_unflatten(
         elif key in nodedef.subgraphs:
           # if the key is a subgraph we create an empty node
           subgraphdef = nodedef.subgraphs[key]
-          if isinstance(subgraphdef, int):
+          if isinstance(subgraphdef, NodeRef):
             # subgraph exists, take it from the cache
-            children[key] = index_to_ref[subgraphdef]
+            children[key] = index_ref[subgraphdef.index]
           else:
             # create a node from an empty state, reasoning:
             # * its a node with no state
@@ -533,13 +521,13 @@ def _graph_unflatten(
             #   created nodes
             substate = {}
             children[key] = _graph_unflatten(
-              subgraphdef, substate, index_to_ref, idxmap
+              subgraphdef, substate, index_ref, index_ref_cache
             )
         elif key in nodedef.leaves:
-          leaf_index = nodedef.leaves[key]
-          if leaf_index is not None and leaf_index in index_to_ref:
+          noderef = nodedef.leaves[key]
+          if noderef is not None and noderef.index in index_ref:
             # variable exists, take it from the cache
-            children[key] = index_to_ref[leaf_index]
+            children[key] = index_ref[noderef.index]
           else:
             # key for a variable is missing, raise an error
             raise ValueError(
@@ -563,29 +551,29 @@ def _graph_unflatten(
           assert isinstance(value, dict)
           subgraphdef = nodedef.subgraphs[key]
 
-          if isinstance(subgraphdef, int):
-            children[key] = index_to_ref[subgraphdef]
+          if isinstance(subgraphdef, NodeRef):
+            children[key] = index_ref[subgraphdef.index]
           else:
             children[key] = _graph_unflatten(
-              subgraphdef, value, index_to_ref, idxmap
+              subgraphdef, value, index_ref, index_ref_cache
             )
 
         elif key in nodedef.leaves:
           if not is_state_leaf(value):
             raise ValueError(f'Expected a leaf for {key!r}, but got {value!r}')
 
-          leaf_index = nodedef.leaves[key]
+          noderef = nodedef.leaves[key]
 
-          if leaf_index is None:
+          if noderef is None:
             # if the leaf is None, it means that the value was originally
             # a non-VariableState leaf, however we allow providing a
             # VariableState presumbly created by modifying the State
             if isinstance(value, VariableState):
               value = value.to_variable()
             children[key] = value
-          elif leaf_index in index_to_ref:
+          elif noderef.index in index_ref:
             # add an existing variable
-            children[key] = index_to_ref[leaf_index]
+            children[key] = index_ref[noderef.index]
           else:
             # its a unseen variable, create a new one
             if not isinstance(value, VariableState):
@@ -594,8 +582,8 @@ def _graph_unflatten(
               )
             # when idxmap is present, check if the Varable exists there
             # and update existing variables if it does
-            if idxmap is not None and leaf_index in idxmap:
-              variable = idxmap[leaf_index]
+            if index_ref_cache is not None and noderef.index in index_ref_cache:
+              variable = index_ref_cache[noderef.index]
               if not isinstance(variable, Variable):
                 raise ValueError(
                   f'Expected a Variable type for {key!r}, but got {type(variable)}.'
@@ -605,7 +593,7 @@ def _graph_unflatten(
               assert isinstance(value, VariableState)
               variable = value.to_variable()
             children[key] = variable
-            index_to_ref[leaf_index] = variable
+            index_ref[noderef.index] = variable
         else:
           raise RuntimeError(f'Unknown key: {key!r}, this is a bug.')
 
@@ -614,8 +602,8 @@ def _graph_unflatten(
   if isinstance(node_impl, GraphNodeImpl):
     # we create an empty node first and add it to the index
     # this avoids infinite recursion when there is a reference cycle
-    if idxmap is not None and nodedef.index in idxmap:
-      node = idxmap[nodedef.index]
+    if index_ref_cache is not None and nodedef.index in index_ref_cache:
+      node = index_ref_cache[nodedef.index]
       if type(node) != nodedef.type:
         raise ValueError(
           f'Expected a node of type {nodedef.type} for index '
@@ -624,7 +612,7 @@ def _graph_unflatten(
       node_impl.clear(node)
     else:
       node = node_impl.create_empty(nodedef.metadata)
-    index_to_ref[nodedef.index] = node
+    index_ref[nodedef.index] = node
     children = _get_children()
     node_impl.init(node, tuple(children.items()))
   else:
@@ -857,35 +845,29 @@ def _graph_update_static(
 # UpdateContext
 # --------------------------------------------------------
 
+@dataclasses.dataclass
+class GraphContext(threading.local):
+  update_context_stacks: dict[str, list[UpdateContext]] = dataclasses.field(
+    default_factory=dict
+  )
+  ref_index_stack: list[SplitContext] = dataclasses.field(default_factory=list)
+  index_ref_stack: list[MergeContext] = dataclasses.field(default_factory=list)
 
-# --------------------------------------------------------
-# UpdateContext
-# --------------------------------------------------------
+
+GRAPH_CONTEXT = GraphContext()
 
 
 @dataclasses.dataclass
-class UpdateContext:
-  """A context manager for handling complex state updates."""
-
-  tag: str
-  refmap: RefMap[tp.Any, Index] | None
-  idxmap: dict[Index, tp.Any] | None
-
-  # define hash and eq to make this an opaque object
-  def __hash__(self):
-    return 0
-
-  def __eq__(self, other):
-    return isinstance(other, UpdateContext)
+class SplitContext:
+  ctxtag: str | None
+  ref_index: RefMap[tp.Any, Index]
 
   @tp.overload
   def split(self, graph_node: A, /) -> tuple[GraphDef[A], GraphState]: ...
-
   @tp.overload
   def split(
     self, graph_node: A, first: filterlib.Filter, /
   ) -> tuple[GraphDef[A], GraphState]: ...
-
   @tp.overload
   def split(
     self,
@@ -895,7 +877,135 @@ class UpdateContext:
     /,
     *filters: filterlib.Filter,
   ) -> tuple[GraphDef[A], GraphState, tpe.Unpack[tuple[GraphState, ...]]]: ...
+  def split(
+    self, node: A, *filters: filterlib.Filter
+  ) -> tuple[GraphDef[A], tpe.Unpack[tuple[GraphState, ...]]]:
+    ctx = (
+      current_update_context(self.ctxtag) if self.ctxtag is not None else None
+    )
+    graphdef, state = flatten(node, self.ref_index)
+    states = _split_state(state, filters)
+    if ctx is not None:
+      if ctx.index_ref is not None and isinstance(graphdef, NodeDef):
+        index_to_index = compose_mapping(ctx.index_ref, self.ref_index)
+        graphdef = dataclasses.replace(
+          graphdef, index_mapping=FrozenDict(index_to_index)
+        )
 
+    return graphdef, *states
+
+
+@contextlib.contextmanager
+def split_context(ctxtag: str | None = None):
+  index_ref: RefMap[tp.Any, Index] = RefMap()
+  flatten_ctx = SplitContext(ctxtag, index_ref)
+  GRAPH_CONTEXT.ref_index_stack.append(flatten_ctx)
+
+  try:
+    yield flatten_ctx
+  finally:
+    GRAPH_CONTEXT.ref_index_stack.pop()
+    if ctxtag is not None:
+      ctx = current_update_context(ctxtag)
+      ctx.flatten_end(index_ref)
+    del flatten_ctx.ref_index
+    del flatten_ctx.ctxtag
+
+
+@dataclasses.dataclass
+class MergeContext:
+  ctxtag: str | None
+  index_ref: dict[Index, tp.Any]
+
+  def merge(
+    self, graphdef: GraphDef[A], state: GraphState, /, *states: GraphState
+  ) -> A:
+    ctx = (
+      current_update_context(self.ctxtag) if self.ctxtag is not None else None
+    )
+    if (
+      ctx is not None
+      and isinstance(graphdef, NodeDef)
+      and graphdef.index_mapping is not None
+    ):
+      # outer merge (4), create index_ref_cache
+      assert ctx.ref_index is not None
+      index_ref_cache = compose_mapping_reversed(
+        ctx.ref_index, graphdef.index_mapping
+      )
+    else:
+      # inner merge (2)
+      index_ref_cache = None
+
+    state = State.merge(state, *states)
+    node = unflatten(
+      graphdef,
+      state,
+      index_ref=self.index_ref,
+      index_ref_cache=index_ref_cache,
+    )
+    return node
+
+
+@contextlib.contextmanager
+def merge_context(ctxtag: str | None = None):
+  index_ref: dict[Index, tp.Any] = {}
+
+  unflatten_ctx = MergeContext(ctxtag, index_ref)
+  GRAPH_CONTEXT.index_ref_stack.append(unflatten_ctx)
+
+  try:
+    yield unflatten_ctx
+  finally:
+    GRAPH_CONTEXT.index_ref_stack.pop()
+    if ctxtag is not None:
+      ctx = current_update_context(ctxtag)
+      ctx.unflatten_end(index_ref)
+    del unflatten_ctx.index_ref
+    del unflatten_ctx.ctxtag
+
+
+@dataclasses.dataclass
+class UpdateContext:
+  """A context manager for handling complex state updates."""
+
+  tag: str
+  ref_index: RefMap[tp.Any, Index] | None
+  index_ref: dict[Index, tp.Any] | None
+
+  # define hash and eq to make this an opaque object
+  def __hash__(self):
+    return 0
+
+  def __eq__(self, other):
+    return isinstance(other, UpdateContext)
+
+  def flatten_end(self, ref_index: RefMap[tp.Any, Index]):
+    if self.ref_index is None:
+      # outer split (1), store the references
+      self.ref_index = ref_index
+    else:
+      # inner split (3), clear index_ref
+      self.index_ref = None
+
+  def unflatten_end(self, index_ref: dict[Index, tp.Any]):
+    self.index_ref = index_ref
+
+  @tp.overload
+  def split(self, graph_node: A, /) -> tuple[GraphDef[A], GraphState]: ...
+  @tp.overload
+  def split(
+    self, graph_node: A, first: filterlib.Filter, /
+  ) -> tuple[GraphDef[A], GraphState]: ...
+  @tp.overload
+  def split(
+    self,
+    graph_node: A,
+    first: filterlib.Filter,
+    second: filterlib.Filter,
+    /,
+    *filters: filterlib.Filter,
+  ) -> tuple[GraphDef[A], GraphState, tpe.Unpack[tuple[GraphState, ...]]]: ...
   def split(
     self, node: A, *filters: filterlib.Filter
   ) -> tuple[GraphDef[A], GraphState, tpe.Unpack[tuple[GraphState, ...]]]:
@@ -907,7 +1017,7 @@ class UpdateContext:
 
     Example usage::
 
-      >>> from flax.experimental import nnx
+      >>> from flax import nnx
       >>> import jax, jax.numpy as jnp
       ...
       >>> class Foo(nnx.Module):
@@ -962,28 +1072,19 @@ class UpdateContext:
       :class:`GraphDef` and one or more :class:`State`'s equal to the number of filters passed. If no
       filters are passed, a single :class:`State` is returned.
     """
-    if self.refmap is not None and self.idxmap is None:
-      raise ValueError(
-        "'merge' was not called in-between the first and second call to 'split'"
+    ref_index: RefMap[tp.Any, Index] = RefMap()
+    graphdef, state = flatten(node, ref_index)
+    states = _split_state(state, filters)
+
+    if self.index_ref is not None and isinstance(graphdef, NodeDef):
+      index_to_index = compose_mapping(self.index_ref, ref_index)
+      graphdef = dataclasses.replace(
+        graphdef, index_mapping=FrozenDict(index_to_index)
       )
-    graphdef, state, refmap = flatten(node, idxmap=self.idxmap)
 
-    states: GraphState | tuple[GraphState, ...]
-    if len(filters) == 0:
-      states = (state,)
-    elif len(filters) == 1:
-      states = (state.split(filters[0]),)
-    else:
-      states = state.split(filters[0], filters[1], *filters[2:])
+    self.flatten_end(ref_index)
 
-    if self.refmap is None:
-      self.refmap = refmap
-
-    if graphdef.index_mapping is not None:
-      # clear idxmap to remove any references to tracers
-      self.idxmap = None
-
-    return graphdef, states[0], *states[1:]
+    return graphdef, *states
 
   def merge(
     self,
@@ -992,22 +1093,30 @@ class UpdateContext:
     *states: GraphState,
   ) -> A:
     """merge"""
-    if self.refmap is None:
-      raise ValueError('Cannot update a graphdef without refmap.')
-
-    if states:
-      state = GraphState.merge(state, *states)
-
-    if graphdef.index_mapping is None:
-      node, self.idxmap = unflatten(graphdef, state)
-    else:
-      index_to_ref = compose_mapping_reversed(
-        self.refmap, graphdef.index_mapping
+    if not isinstance(graphdef, NodeDef):
+      raise ValueError(
+        f'Expected a NodeDef instance, but got {type(graphdef)}.'
       )
-      node, _idxmap = unflatten(graphdef, state, idxmap=index_to_ref)
-      # clear references
-      self.refmap = None
-      self.idxmap = None
+    if self.ref_index is None:
+      raise ValueError('Cannot merge without ref_index.')
+
+    if graphdef.index_mapping is not None:
+      # outer merge (4), create index_ref_cache
+      assert self.ref_index is not None
+      index_ref_cache = compose_mapping_reversed(
+        self.ref_index, graphdef.index_mapping
+      )
+    else:
+      # inner merge (2)
+      index_ref_cache = None
+
+    state = State.merge(state, *states)
+    index_ref: dict[Index, tp.Any] = {}
+    node = unflatten(
+      graphdef, state, index_ref=index_ref, index_ref_cache=index_ref_cache
+    )
+
+    self.unflatten_end(index_ref)
 
     return node
 
@@ -1021,20 +1130,26 @@ class UpdateContextManager:
 
   def __enter__(self):
     ctx = UpdateContext(self.tag, None, None)
-    GRAPH_CONTEXT.update_context_stacks[self.tag].append(ctx)
+    if self.tag not in GRAPH_CONTEXT.update_context_stacks:
+      GRAPH_CONTEXT.update_context_stacks[self.tag] = [ctx]
+    else:
+      GRAPH_CONTEXT.update_context_stacks[self.tag].append(ctx)
     return ctx
 
   def __exit__(self, *args):
-    stack = GRAPH_CONTEXT.update_context_stacks[self.tag]
-    if not stack:
+    if self.tag not in GRAPH_CONTEXT.update_context_stacks:
       raise RuntimeError(
           f'No update context found for tag {self.tag!r}, this is a bug.'
       )
+    stack = GRAPH_CONTEXT.update_context_stacks[self.tag]
 
-    ctx = GRAPH_CONTEXT.update_context_stacks[self.tag].pop()
+    ctx = stack.pop()
     # clear references
-    ctx.refmap = None
-    ctx.idxmap = None
+    del ctx.ref_index
+    del ctx.index_ref
+
+    if not stack:
+      del GRAPH_CONTEXT.update_context_stacks[self.tag]
 
   def __call__(self, f: F) -> F:
     @functools.wraps(f)
@@ -1142,29 +1257,34 @@ def update_context(tag: str):
 
 def current_update_context(tag: str) -> UpdateContext:
   """Returns the current active :class:`UpdateContext` for the given tag."""
-  stack = GRAPH_CONTEXT.update_context_stacks[tag]
-  if not stack:
+  if tag not in GRAPH_CONTEXT.update_context_stacks:
     raise ValueError(f'No update context found for tag {tag!r}.')
-  return stack[-1]
+  return GRAPH_CONTEXT.update_context_stacks[tag][-1]
 
 
 # --------------------------------------------------------
 # Functional API
 # --------------------------------------------------------
 
+def _split_state(
+  state: GraphState,
+  filters: tuple[filterlib.Filter, ...],
+) -> tuple[GraphState, tpe.Unpack[tuple[GraphState, ...]]]:
+  if not filters:
+    return (state,)
+  states = state.split(*filters)
+  if isinstance(states, State):
+    return (states,)
+  assert len(states) > 0
+  return states  # type: ignore[return-value]
+
 
 @tp.overload
 def split(graph_node: A, /) -> tuple[GraphDef[A], GraphState]: ...
-
-
 @tp.overload
 def split(
-  graph_node: A,
-  first: filterlib.Filter,
-  /,
+  graph_node: A, first: filterlib.Filter, /
 ) -> tuple[GraphDef[A], GraphState]: ...
-
-
 @tp.overload
 def split(
   graph_node: A,
@@ -1173,8 +1293,6 @@ def split(
   /,
   *filters: filterlib.Filter,
 ) -> tuple[GraphDef[A], GraphState, tpe.Unpack[tuple[GraphState, ...]]]: ...
-
-
 def split(
   node: A, *filters: filterlib.Filter
 ) -> tuple[GraphDef[A], GraphState, tpe.Unpack[tuple[GraphState, ...]]]:
@@ -1186,7 +1304,7 @@ def split(
 
   Example usage::
 
-    >>> from flax.experimental import nnx
+    >>> from flax import nnx
     >>> import jax, jax.numpy as jnp
     ...
     >>> class Foo(nnx.Module):
@@ -1246,17 +1364,9 @@ def split(
     ``GraphDef`` and one or more ``States`` equal to the number of filters passed. If no
     filters are passed, a single ``State`` is returned.
   """
-  graphdef, state, _ = flatten(node)
-
-  states: GraphState | tuple[GraphState, ...]
-  if len(filters) == 0:
-    states = (state,)
-  elif len(filters) == 1:
-    states = (state.split(filters[0]),)
-  else:
-    states = state.split(filters[0], filters[1], *filters[2:])
-
-  return graphdef, states[0], *states[1:]
+  graphdef, state = flatten(node)
+  states = _split_state(state, filters)
+  return graphdef, *states
 
 
 def merge(
@@ -1272,7 +1382,7 @@ def merge(
 
   Example usage::
 
-    >>> from flax.experimental import nnx
+    >>> from flax import nnx
     >>> import jax, jax.numpy as jnp
     ...
     >>> class Foo(nnx.Module):
@@ -1300,10 +1410,8 @@ def merge(
   Returns:
     The merged :class:`Module`.
   """
-  if states:
-    state = GraphState.merge(state, *states)
-
-  node, _ = unflatten(graphdef, state)
+  state = GraphState.merge(state, *states)
+  node = unflatten(graphdef, state)
   return node
 
 
@@ -1390,7 +1498,7 @@ def state(
   Returns:
     One or more :class:`State` mappings.
   """
-  state = flatten(node)[1]
+  _, state = flatten(node)
 
   states: GraphState | tuple[GraphState, ...]
   if len(filters) == 0:
@@ -1419,7 +1527,7 @@ def graphdef(node: tp.Any, /) -> GraphDef[tp.Any]:
   Returns:
     The :class:`GraphDef` of the :class:`Module` object.
   """
-  graphdef, _, _ = flatten(node)
+  graphdef, _ = flatten(node)
   return graphdef
 
 
@@ -1523,6 +1631,96 @@ def clone(node: Node) -> Node:
   return merge(graphdef, state)
 
 
+def call(
+  graphdef_state: tuple[GraphDef[A], GraphState], /
+) -> ApplyCaller[tuple[GraphDef[A], GraphState]]:
+  """Calls a method underlying graph node defined by a (GraphDef, State) pair.
+
+  ``call`` takes a ``(GraphDef, State)`` pair and creates a proxy object that can be
+  used to call methods on the underlying graph node. When a method is called, the
+  output is returned along with a new (GraphDef, State) pair that represents the
+  updated state of the graph node. ``call`` is equivalent to :func:`merge` > ``method``
+  > :func:`split`` but is more convenient to use in pure JAX functions.
+
+  Example::
+
+    >>> from flax import nnx
+    >>> import jax
+    >>> import jax.numpy as jnp
+    ...
+    >>> class StatefulLinear(nnx.Module):
+    ...   def __init__(self, din, dout, rngs):
+    ...     self.w = nnx.Param(jax.random.uniform(rngs(), (din, dout)))
+    ...     self.b = nnx.Param(jnp.zeros((dout,)))
+    ...     self.count = nnx.Variable(jnp.array(0, dtype=jnp.uint32))
+    ...
+    ...   def increment(self):
+    ...     self.count += 1
+    ...
+    ...   def __call__(self, x):
+    ...     self.increment()
+    ...     return x @ self.w + self.b
+    ...
+    >>> linear = StatefulLinear(3, 2, nnx.Rngs(0))
+    >>> linear_state = nnx.split(linear)
+    ...
+    >>> @jax.jit
+    ... def forward(x, linear_state):
+    ...   y, linear_state = nnx.call(linear_state)(x)
+    ...   return y, linear_state
+    ...
+    >>> x = jnp.ones((1, 3))
+    >>> y, linear_state = forward(x, linear_state)
+    >>> y, linear_state = forward(x, linear_state)
+    ...
+    >>> linear = nnx.merge(*linear_state)
+    >>> linear.count.value
+    Array(2, dtype=uint32)
+
+  The proxy object returned by ``call`` supports indexing and attribute access
+  to access nested methods. In the example below, the ``increment`` method indexing
+  is used to call the ``increment`` method of the ``StatefulLinear`` module
+  at the ``b`` key of a ``nodes`` dictionary.
+
+    >>> class StatefulLinear(nnx.Module):
+    ...   def __init__(self, din, dout, rngs):
+    ...     self.w = nnx.Param(jax.random.uniform(rngs(), (din, dout)))
+    ...     self.b = nnx.Param(jnp.zeros((dout,)))
+    ...     self.count = nnx.Variable(jnp.array(0, dtype=jnp.uint32))
+    ...
+    ...   def increment(self):
+    ...     self.count += 1
+    ...
+    ...   def __call__(self, x):
+    ...     self.increment()
+    ...     return x @ self.w + self.b
+    ...
+    >>> rngs = nnx.Rngs(0)
+    >>> nodes = dict(
+    ...   a=StatefulLinear(3, 2, rngs),
+    ...   b=StatefulLinear(2, 1, rngs),
+    ... )
+    ...
+    >>> node_state = nnx.split(nodes)
+    >>> # use attribute access
+    >>> _, node_state = nnx.call(node_state)['b'].increment()
+    ...
+    >>> nodes = nnx.merge(*node_state)
+    >>> nodes['a'].count.value
+    Array(0, dtype=uint32)
+    >>> nodes['b'].count.value
+    Array(1, dtype=uint32)
+  """
+
+  def pure_caller(accessor: DelayedAccessor, *args, **kwargs):
+    node = merge(*graphdef_state)
+    method = accessor(node)
+    out = method(*args, **kwargs)
+    return out, split(node)
+
+  return CallableProxy(pure_caller)  # type: ignore
+
+
 def iter_graph(node: tp.Any, /) -> tp.Iterator[tuple[PathParts, tp.Any]]:
   """Iterates over all nested nodes and leaves of the given graph node, including the current node.
 
@@ -1595,50 +1793,6 @@ class Static(tp.Generic[A]):
 
 jax.tree_util.register_static(Static)
 
-# ---------------------------------------------------------
-# insert/extract_graph_nodes API
-# ---------------------------------------------------------
-
-
-@dataclasses.dataclass(frozen=True)
-class GraphNodeIndex:
-  """Index of a graph node in a Pytree structure."""
-
-  index: Index
-
-
-jax.tree_util.register_static(GraphNodeIndex)
-
-
-def extract_graph_nodes(pytree: A, /) -> tuple[A, tuple[tp.Any, ...]]:
-  """Extracts all graph nodes from a pytree."""
-  nodes = RefMap[tp.Any, Index]()
-
-  def _maybe_extract(x):
-    if is_graph_node(x):
-      if x not in nodes:
-        index = nodes[x] = len(nodes)
-      else:
-        index = nodes[x]
-      return GraphNodeIndex(index)
-    return x
-
-  return jax.tree_util.tree_map(_maybe_extract, pytree), tuple(nodes)
-
-
-def insert_graph_nodes(pytree: A, nodes: tuple[tp.Any, ...], /) -> A:
-  """Inserts graph nodes into a pytree."""
-
-  def _maybe_insert(x):
-    if isinstance(x, GraphNodeIndex):
-      return nodes[x.index]
-    return x
-
-  return jax.tree_util.tree_map(
-    _maybe_insert, pytree, is_leaf=lambda x: isinstance(x, GraphNodeIndex)
-  )
-
-
 # ---------------------------------------------------------
 # Pytree
 # ---------------------------------------------------------
@@ -1646,7 +1800,7 @@ class PytreeType: ...
 
 
 def is_pytree_node(x: tp.Any) -> bool:
-  return not jax.tree_util.all_leaves([x])
+  return not jax.tree_util.all_leaves((x,))
 
 
 def _key_path_to_key(key: tp.Any) -> Key:
