@@ -50,7 +50,7 @@ AxisIndex = int
 AddAxisHook = tp.Callable[[V, AxisName, AxisIndex], None]
 RemoveAxisHook = tp.Callable[[V, AxisName, AxisIndex], None]
 
-VariableTypeCache: dict[str, tp.Type['Variable[tp.Any]']] = {}
+VariableTypeCache: dict[str, tp.Type[Variable[tp.Any]]] = {}
 
 
 class Empty:
@@ -86,8 +86,8 @@ class VariableMetadata(tp.Generic[A]):
   set_value_hooks: tuple[SetValueHook[A], ...] = ()
   get_value_hooks: tuple[GetValueHook[A], ...] = ()
   create_value_hooks: tuple[CreateValueHook[A], ...] = ()
-  add_axis_hooks: tuple[AddAxisHook['Variable[A]'], ...] = ()
-  remove_axis_hooks: tuple[RemoveAxisHook['Variable[A]'], ...] = ()
+  add_axis_hooks: tuple[AddAxisHook[Variable[A]], ...] = ()
+  remove_axis_hooks: tuple[RemoveAxisHook[Variable[A]], ...] = ()
   metadata: tp.Mapping[str, tp.Any] = dataclasses.field(default_factory=dict)
 
 
@@ -160,8 +160,8 @@ class Variable(tp.Generic[A], reprlib.Representable):
   set_value_hooks: tuple[SetValueHook[A], ...]
   get_value_hooks: tuple[GetValueHook[A], ...]
   create_value_hooks: tuple[CreateValueHook[A], ...]
-  add_axis_hooks: tuple[AddAxisHook['Variable[A]'], ...]
-  remove_axis_hooks: tuple[RemoveAxisHook['Variable[A]'], ...]
+  add_axis_hooks: tuple[AddAxisHook[Variable[A]], ...]
+  remove_axis_hooks: tuple[RemoveAxisHook[Variable[A]], ...]
   _trace_state: tracers.TraceState
 
   def __init__(
@@ -177,11 +177,11 @@ class Variable(tp.Generic[A], reprlib.Representable):
       CreateValueHook[A], tp.Sequence[CreateValueHook[A]]
     ] = (),
     add_axis_hooks: tp.Union[
-      AddAxisHook['Variable[A]'], tp.Sequence[AddAxisHook['Variable[A]']]
+      AddAxisHook[Variable[A]], tp.Sequence[AddAxisHook[Variable[A]]]
     ] = (),
     remove_axis_hooks: tp.Union[
-      RemoveAxisHook['Variable[A]'],
-      tp.Sequence[RemoveAxisHook['Variable[A]']],
+      RemoveAxisHook[Variable[A]],
+      tp.Sequence[RemoveAxisHook[Variable[A]]],
     ] = (),
     **metadata: tp.Any,
   ):
@@ -304,10 +304,10 @@ class Variable(tp.Generic[A], reprlib.Representable):
     object.__setattr__(self, name, value)
 
   @classmethod
-  def state(cls, value: A, **metadata) -> 'VariableState[A]':
+  def state(cls, value: A, **metadata) -> VariableState[A]:
     return cls(value, **metadata).to_state()
 
-  def copy_from(self, other: 'Variable[A]') -> None:
+  def copy_from(self, other: Variable[A]) -> None:
     if type(self) is not type(other):
       raise ValueError(
         f'Cannot copy from incompatible container, '
@@ -322,7 +322,7 @@ class Variable(tp.Generic[A], reprlib.Representable):
     vars_dict.clear()
     vars_dict.update(other_vars, _trace_state=trace_state)
 
-  def copy_from_state(self, variable_state: 'VariableState[A]'):
+  def copy_from_state(self, variable_state: VariableState[A]):
     trace_state = self._trace_state
     variable_vars = vars(self)
     variable_vars.clear()
@@ -368,12 +368,12 @@ class Variable(tp.Generic[A], reprlib.Representable):
     return type(self) is type(other) and vars(other) == vars(self)
 
   @tp.overload
-  def replace(self, value: B, **kwargs) -> 'Variable[B]': ...
+  def replace(self, value: B, **kwargs) -> Variable[B]: ...
 
   @tp.overload
-  def replace(self, **kwargs) -> 'Variable[A]': ...
+  def replace(self, **kwargs) -> Variable[A]: ...
 
-  def replace(self, value: tp.Any = MISSING, **kwargs) -> 'Variable[tp.Any]':
+  def replace(self, value: tp.Any = MISSING, **kwargs) -> Variable[tp.Any]:
     if value is not MISSING:
       kwargs['raw_value'] = value
 
@@ -407,14 +407,14 @@ class Variable(tp.Generic[A], reprlib.Representable):
     vars(obj).update(attributes)
     return obj
 
-  def copy(self: 'Variable[A]') -> 'Variable[A]':
+  def copy(self: Variable[A]) -> Variable[A]:
     obj = object.__new__(type(self))
     attributes = vars(self).copy()
     attributes['_trace_state'] = tracers.TraceState()
     vars(obj).update(attributes)
     return obj
 
-  def to_state(self: 'Variable[A]') -> 'VariableState[A]':
+  def to_state(self: Variable[A]) -> VariableState[A]:
     metadata = vars(self).copy()
     del metadata['raw_value']
     del metadata['_trace_state']
@@ -429,8 +429,8 @@ class Variable(tp.Generic[A], reprlib.Representable):
         continue
       yield reprlib.Attr(name, repr(value))
 
-  def __penzai_repr__(self, path, subtree_renderer):
-    from penzai.treescope import repr_lib as pz_repr_lib  # type: ignore[import-not-found,import-untyped]
+  def __treescope_repr__(self, path, subtree_renderer):
+    import treescope  # type: ignore[import-not-found,import-untyped]
     children = {}
     for name, value in vars(self).items():
       if name == 'raw_value':
@@ -438,7 +438,7 @@ class Variable(tp.Generic[A], reprlib.Representable):
       if name.endswith('_hooks') or name == '_trace_state':
         continue
       children[name] = value
-    return pz_repr_lib.render_object_constructor(
+    return treescope.repr_lib.render_object_constructor(
         object_type=type(self),
         attributes=children,
         path=path,
@@ -465,7 +465,15 @@ class Variable(tp.Generic[A], reprlib.Representable):
     ) -> V:
       raise NotImplementedError
 
-  # operator overloads
+  # overloads
+  @property
+  def shape(self) -> tuple[int, ...]:
+    return self.value.shape  # type: ignore
+
+  @property
+  def dtype(self) -> Any:
+    return self.value.dtype  # type: ignore
+
   def __jax_array__(self):
     return self.value
 
@@ -556,44 +564,109 @@ class Variable(tp.Generic[A], reprlib.Representable):
   def __ror__(self, other) -> A:
     return self.value.__ror__(other)  # type: ignore
 
-  def __iadd__(self, other) -> A:
-    return self.value.__iadd__(other)  # type: ignore
+  def __iadd__(self: V, other) -> V:
+    value = self.value
+    if hasattr(value, '__iadd__'):
+      value.__iadd__(other)
+    else:
+      self.value = value.__add__(other)
+    return self
 
-  def __isub__(self, other) -> A:
-    return self.value.__isub__(other)  # type: ignore
+  def __isub__(self: V, other) -> V:
+    value = self.value
+    if hasattr(value, '__isub__'):
+      value.__isub__(other)
+    else:
+      self.value = value.__sub__(other)
+    return self
 
-  def __imul__(self, other) -> A:
-    return self.value.__imul__(other)  # type: ignore
+  def __imul__(self: V, other) -> V:
+    value = self.value
+    if hasattr(value, '__imul__'):
+      value.__imul__(other)
+    else:
+      self.value = value.__mul__(other)
+    return self
 
-  def __imatmul__(self, other) -> A:
-    return self.value.__imatmul__(other)  # type: ignore
+  def __imatmul__(self: V, other) -> V:
+    value = self.value
+    if hasattr(value, '__imatmul__'):
+      value.__imatmul__(other)
+    else:
+      self.value = value.__matmul__(other)
+    return self
 
-  def __itruediv__(self, other) -> A:
-    return self.value.__itruediv__(other)  # type: ignore
+  def __itruediv__(self: V, other) -> V:
+    value = self.value
+    if hasattr(value, '__itruediv__'):
+      value.__itruediv__(other)
+    else:
+      self.value = value.__truediv__(other)
+    return self
 
-  def __ifloordiv__(self, other) -> A:
-    return self.value.__ifloordiv__(other)  # type: ignore
+  def __ifloordiv__(self: V, other) -> V:
+    value = self.value
+    if hasattr(value, '__ifloordiv__'):
+      value.__ifloordiv__(other)
+    else:
+      self.value = value.__floordiv__(other)
+    return self
 
-  def __imod__(self, other) -> A:
-    return self.value.__imod__(other)  # type: ignore
+  def __imod__(self: V, other) -> V:
+    value = self.value
+    if hasattr(value, '__imod__'):
+      value.__imod__(other)
+    else:
+      self.value = value.__mod__(other)
+    return self
 
-  def __ipow__(self, other) -> A:
-    return self.value.__ipow__(other)  # type: ignore
+  def __ipow__(self: V, other) -> V:
+    value = self.value
+    if hasattr(value, '__ipow__'):
+      value.__ipow__(other)
+    else:
+      self.value = value.__pow__(other)
+    return self
 
-  def __ilshift__(self, other) -> A:
-    return self.value.__ilshift__(other)  # type: ignore
+  def __ilshift__(self: V, other) -> V:
+    value = self.value
+    if hasattr(value, '__ilshift__'):
+      value.__ilshift__(other)
+    else:
+      self.value = value.__lshift__(other)
+    return self
 
-  def __irshift__(self, other) -> A:
-    return self.value.__irshift__(other)  # type: ignore
+  def __irshift__(self: V, other) -> V:
+    value = self.value
+    if hasattr(value, '__irshift__'):
+      value.__irshift__(other)
+    else:
+      self.value = value.__rshift__(other)
+    return self
 
-  def __iand__(self, other) -> A:
-    return self.value.__iand__(other)  # type: ignore
+  def __iand__(self: V, other) -> V:
+    value = self.value
+    if hasattr(value, '__iand__'):
+      value.__iand__(other)
+    else:
+      self.value = value.__and__(other)
+    return self
 
-  def __ixor__(self, other) -> A:
-    return self.value.__ixor__(other)  # type: ignore
+  def __ixor__(self: V, other) -> V:
+    value = self.value
+    if hasattr(value, '__ixor__'):
+      value.__ixor__(other)
+    else:
+      self.value = value.__xor__(other)
+    return self
 
-  def __ior__(self, other) -> A:
-    return self.value.__ior__(other)  # type: ignore
+  def __ior__(self: V, other) -> V:
+    value = self.value
+    if hasattr(value, '__ior__'):
+      value.__ior__(other)
+    else:
+      self.value = value.__or__(other)
+    return self
 
   def __neg__(self) -> A:
     return self.value.__neg__()  # type: ignore
@@ -780,21 +853,21 @@ class VariableState(tp.Generic[A], reprlib.Representable):
         continue
       yield reprlib.Attr(name, repr(value))
 
-  def __penzai_repr__(self, path, subtree_renderer):
-    from penzai.treescope import repr_lib as pz_repr_lib  # type: ignore[import-not-found,import-untyped]
+  def __treescope_repr__(self, path, subtree_renderer):
+    import treescope  # type: ignore[import-not-found,import-untyped]
     children = {'type': self.type}
     for name, value in vars(self).items():
       if name == 'type' or name.endswith('_hooks'):
         continue
       children[name] = value
-    return pz_repr_lib.render_object_constructor(
+    return treescope.repr_lib.render_object_constructor(
         object_type=type(self),
         attributes=children,
         path=path,
         subtree_renderer=subtree_renderer,
     )
 
-  def replace(self, value: B) -> 'VariableState[B]':
+  def replace(self, value: B) -> VariableState[B]:
     return VariableState(self.type, value, **self.get_metadata())
 
   def to_variable(self) -> Variable[A]:
@@ -829,7 +902,7 @@ class VariableState(tp.Generic[A], reprlib.Representable):
 def _variable_state_flatten(x: VariableState[tp.Any], *, with_keys: bool):
   metadata = tuple(x.get_metadata().items())
   if with_keys:
-    node = (jtu.GetAttrKey('raw_value'), x.value)
+    node = (jtu.GetAttrKey('value'), x.value)
   else:
     node = x.value
 
@@ -863,11 +936,11 @@ def with_metadata(
     CreateValueHook[A], tp.Sequence[CreateValueHook[A]]
   ] = (),
   add_axis_hooks: tp.Union[
-    AddAxisHook['Variable[A]'], tp.Sequence[AddAxisHook['Variable[A]']]
+    AddAxisHook[Variable[A]], tp.Sequence[AddAxisHook[Variable[A]]]
   ] = (),
   remove_axis_hooks: tp.Union[
-    RemoveAxisHook['Variable[A]'],
-    tp.Sequence[RemoveAxisHook['Variable[A]']],
+    RemoveAxisHook[Variable[A]],
+    tp.Sequence[RemoveAxisHook[Variable[A]]],
   ] = (),
   **metadata: tp.Any,
 ) -> F:
@@ -926,14 +999,47 @@ def with_metadata(
   return wrapper  # type: ignore
 
 
+### Variable type <-> name mapping ###
+# Assumption: the mapping is 1-1 and unique.
+
 def variable_type(name: str) -> tp.Type[Variable[tp.Any]]:
+  """Given a Linen-style collection name, get or create its corresponding NNX Variable type."""
   if name not in VariableTypeCache:
     VariableTypeCache[name] = type(name, (Variable,), {})
   return VariableTypeCache[name]
 
 
+def variable_type_name(typ: tp.Type[Variable[tp.Any]]) -> str:
+  """Given an NNX Variable type, get or create its Linen-style collection name.
+
+  Should output the exact inversed result of `variable_type()`."""
+  for name, t in VariableTypeCache.items():
+    if typ == t:
+      return name
+  name = typ.__name__
+  if name in VariableTypeCache:
+    raise ValueError(
+      'Name {name} is already registered in the registry as {VariableTypeCache[name]}. '
+      'It cannot be linked with this type {typ}.'
+    )
+  register_variable_name_type_pair(name, typ)
+  return name
+
+
+def register_variable_name_type_pair(name, typ):
+  """Register a pair of variable type name (like Linen collections) and its NNX type."""
+  VariableTypeCache[name] = typ
+
+
 # add known variable type names
-VariableTypeCache['params'] = Param
-VariableTypeCache['batch_stats'] = BatchStat
-VariableTypeCache['cache'] = Cache
-VariableTypeCache['intermediates'] = Intermediate
+register_variable_name_type_pair('params', Param)
+register_variable_name_type_pair('batch_stats', BatchStat)
+register_variable_name_type_pair('cache', Cache)
+register_variable_name_type_pair('intermediates', Intermediate)
+
+
+def sort_variable_types(types: list[type]):
+  def _variable_parents_count(t: type):
+    return sum(1 for p in t.mro() if issubclass(p, nnx.Variable))
+  parent_count = {t: _variable_parents_count(t) for t in types}
+  return sorted(types, key=lambda t: -parent_count[t])
